<?php

/**
 * Running submission states.
 */
define('SIMPLYTEST_STATUS_ENQUEUE',          100);
define('SIMPLYTEST_STATUS_SPAWNED',          110);
define('SIMPLYTEST_STATUS_PREPARE',          120);
define('SIMPLYTEST_STATUS_DOWNLOAD',         130);
define('SIMPLYTEST_STATUS_PATCHING',         140);
define('SIMPLYTEST_STATUS_INSTALLING',       150);
define('SIMPLYTEST_STATUS_FINALIZE',         160);
define('SIMPLYTEST_STATUS_FINISHED',         170);

/**
 * Terminated submission states.
 */
define('SIMPLYTEST_STATUS_TERMINATED',       200);
define('SIMPLYTEST_STATUS_ABORTED',          210);
define('SIMPLYTEST_STATUS_FAILED',           220);

/**
 * Failure submission states. One error for each running state.
 */
define('SIMPLYTEST_STATUS_ERROR_SERVER',     300);
define('SIMPLYTEST_STATUS_ERROR_SPAWNED',    310);
define('SIMPLYTEST_STATUS_ERROR_PREPARE',    320);
define('SIMPLYTEST_STATUS_ERROR_DOWNLOAD',   330);
define('SIMPLYTEST_STATUS_ERROR_PATCHING',   340);
define('SIMPLYTEST_STATUS_ERROR_INSTALLING', 350);
define('SIMPLYTEST_STATUS_ERROR_FINALIZE',   360);

/**
 * Get a list of all submission statuses with their label and messages.
 *
 * @param null|string $type
 *
 * @return array
 */
function simplytest_tugboat_status_list($type = NULL) {
  $statuses = [
    // 100s - Running submission states.
    SIMPLYTEST_STATUS_ENQUEUE => [
      'label' => t('Enqueued'),
      'message' => t('Your submission is enqueued.'),
      'type' => 'running',
    ],
    SIMPLYTEST_STATUS_SPAWNED => [
      'label' => t('Spawned'),
      'message' => t('The submission was spawned.'),
      'type' => 'running',
    ],
    SIMPLYTEST_STATUS_PREPARE => [
      'label' => t('Preparing'),
      'message' => t('The environment is being prepared.'),
      'type' => 'running',
    ],
    SIMPLYTEST_STATUS_DOWNLOAD => [
      'label' => t('Downloading'),
      'message' => t('Fetching and downloading dependencies.'),
      'type' => 'running',
    ],
    SIMPLYTEST_STATUS_PATCHING => [
      'label' => t('Patching'),
      'message' => t('Downloading and applying patches.'),
      'type' => 'running',
    ],
    SIMPLYTEST_STATUS_INSTALLING => [
      'label' => t('Installing'),
      'message' => t('Setup and installation.'),
      'type' => 'running',
    ],
    SIMPLYTEST_STATUS_FINALIZE => [
      'label' => t('Finalizing'),
      'message' => t('Final polish.'),
      'type' => 'running',
    ],
    SIMPLYTEST_STATUS_FINISHED => [
      'label' => t('Finished'),
      'message' => t('Finished!'),
      'type' => 'running',
    ],

    // 200s - Terminated submission states.
    SIMPLYTEST_STATUS_TERMINATED => [
      'label' => t('Terminated'),
      'message' => t('This submission is already terminated.'),
      'type' => 'terminated',
    ],
    SIMPLYTEST_STATUS_ABORTED => [
      'label' => t('Aborted'),
      'message' => t('The requested submission was aborted.'),
      'type' => 'terminated',
    ],
    SIMPLYTEST_STATUS_FAILED => [
      'label' => t('Failed'),
      'message' => t('The requested submission failed.'),
      'type' => 'terminated',
    ],

    // 300s - Failure submission states.
    SIMPLYTEST_STATUS_ERROR_SERVER => [
      'label' => t('Error server'),
      'message' => t('An error occurred while launching the submission.'),
      'type' => 'error',
    ],
    SIMPLYTEST_STATUS_ERROR_SPAWNED => [
      'label' => t('Error spawning'),
      'message' => t('An error occurred while spawning the environment.'),
      'type' => 'error',
    ],
    SIMPLYTEST_STATUS_ERROR_PREPARE => [
      'label' => t('Error prepare'),
      'message' => t('An error occurred while preparing the environment.'),
      'type' => 'error',
    ],
    SIMPLYTEST_STATUS_ERROR_DOWNLOAD => [
      'label' => t('Error download'),
      'message' => t('An error occurred while downloading dependencies.'),
      'type' => 'error',
    ],
    SIMPLYTEST_STATUS_ERROR_PATCHING => [
      'label' => t('Error patching'),
      'message' => t('An error occurred while patching the project.'),
      'type' => 'error',
    ],
    SIMPLYTEST_STATUS_ERROR_INSTALLING => [
      'label' => t('Error installing'),
      'message' => t('An error occurred while installing the environment.'),
      'type' => 'error',
    ],
    SIMPLYTEST_STATUS_ERROR_FINALIZE => [
      'label' => t('Error finalizing'),
      'message' => t('An error occurred while finalizing the environment.'),
      'type' => 'error',
    ],
  ];

  foreach ($statuses as $code => $status) {
    $statuses[$code]['code'] = $code;
  }

  if ($type) {
    $statuses = array_filter($statuses, function ($status) use ($type) {
      return $status['type'] == $type;
    });
  }

  return $statuses;
}

/**
 * Get a specific status by its code.
 *
 * @param $status_coded
 *
 * @return bool|mixed
 */
function simplytest_tugboat_status_get($status_coded) {
  $statuses = simplytest_tugboat_status_list();

  return !empty($statuses[$status_coded]) ? $statuses[$status_coded] : FALSE;
}

/**
 * Page callback for the tugboat progress indicator.
 *
 * @param $id
 *
 * @return array
 */
function simplytest_tugboat_status_progress_page($id) {

  // Make sure this submission is available.
  $state = simplytest_tugboat_status_state($id);
  if ($state === FALSE) {
    drupal_set_message(t('The requested submission is not available.'), 'error');
    drupal_goto();
  }

  // If the siterequest came by the meta no_js refresh.
  if (isset($_GET['no_js'])) {
    // If this submission is finished, HTTP redirect.
    if ($state['percent'] == '100') {
      simplytest_tugboat_status_goto($id);
      return;
    }
  }

  // Check for an established URL, forward if exists.
  if ($state['percent'] == '100') {
    $url = simplytest_tugboat_url_load($id);

    // Redirect through a 'Refresh' meta tag if JavaScript is disabled.
    $meta_refresh = array(
      '#prefix' => '<noscript>',
      '#suffix' => '</noscript>',
      '#tag' => 'meta',
      '#attributes' => array(
        'http-equiv' => 'Refresh',
        'content' => '5; URL=' . $url,
      ),
    );
    drupal_add_html_head($meta_refresh, 'simplytest_tugboat_status_meta_refresh');

    return array(
      '#type' => 'container',
      'page-contents' => array(
        '#type' => 'markup',
        '#markup' => '<p>Forwarding you to the Tugboat instance</p>',
      ),
    );

  } else {

    watchdog('message', 'State: ' . var_export($state, TRUE));

    // Redirect through a 'Refresh' meta tag if JavaScript is disabled.
    $meta_refresh = array(
      '#prefix' => '<noscript>',
      '#suffix' => '</noscript>',
      '#tag' => 'meta',
      '#attributes' => array(
        'http-equiv' => 'Refresh',
        // Keep refreshing the current page, but mark each refresh with no_js.
        'content' => '5; URL=' . url(current_path(), array('query' => array('no_js' => NULL))),
      ),
    );
    drupal_add_html_head($meta_refresh, 'simplytest_tugboat_status_meta_refresh');

    // Return a progress bar and attach own javascript.
    return array(
      '#type' => 'container',
      '#attributes' => array(
          'class' => array('simplytest-progress-bar'),
       ),
      '#attached' => array(
        'js' => array(
          array(
            'data' => array(
              'simplytest_tugboat' => array(
                'id' => $id,
              ),
            ),
            'type' => 'setting',
          ),
        ),
      ),
      'progress-bar' => array(
        '#theme' => 'progress_bar',
        '#percent' => $state['percent'],
        '#message' => $state['message'],
      ),
      'log' => array(
        '#prefix' => '<div id="simplytest-log" class="log">',
        '#markup' => $state['log'],
	'#suffix' => '</div>',
        //'#children' => t('number of items: ' . count($state['log'])),
        '#cache' => array(
          'expire' => time(),
        ),
      ),
    );
  }
}

/**
 * Get all statuses for the given instance.
 *
 * @param $instance_id
 *
 * @return \stdClass|FALSE
 */
function simplytest_tugboat_status_get_instance_statuses($instance_id) {
  $results = db_select('simplytest_tugboat_status', 'sts')
    ->fields('sts')
    ->condition('instance_id', $instance_id)
    ->orderBy('timestamp')
    ->execute()
    ->fetchAll();

  return $results;
}

/**
 * Update instance status.
 *
 * @param $instance_id
 * @param $status
 */
function simplytest_tugboat_status_update($instance_id, $status) {
  $status = (int) $status;
  $status_valid = simplytest_tugboat_status_get($status);

  if (!$status_valid) {
    return;
  }

  $exists = db_select('simplytest_tugboat_status', 'sts')
    ->fields('sts', ['instance_id'])
    ->condition('instance_id', $instance_id)
    ->condition('status', $status)
    ->execute()
    ->fetchCol('instance_id');

  // Don't set the same status twice for an instance.
  if ($exists) {
    return;
  }

  try {
    db_insert('simplytest_tugboat_status')->fields([
      'instance_id' => $instance_id,
      'status' => $status,
      'timestamp' => time(),
    ])->execute();
  }
  catch (\Exception $exception) {
    watchdog_exception('simplytest_tugboat', $exception);
  }
}

/**
 * Callback for remote status update.
 *
 * @param $instance_id
 * @param $status
 */
function simplytest_tugboat_status_update_remote($instance_id, $status) {
  simplytest_tugboat_status_update($instance_id, $status);
  drupal_exit();
}

/**
 * Instance state as array.
 *
 * @param $instance_id
 *
 * @return array
 */
function simplytest_tugboat_status_state($instance_id) {
  $statuses = simplytest_tugboat_status_get_instance_statuses($instance_id);
  $status_code = FALSE;
  $percent = 0;
  $message = t('No instance status found...');
  $log = array();

  if (!empty($statuses)) {
    $instance_status = array_pop($statuses);
    $status = simplytest_tugboat_status_get($instance_status->status);

    if ($status) {
      $status_code = $status['code'];
      $percent = simplytest_tugboat_status_running_percentage($status['code']);
      $message = $status['message'];
      $log = _simplytest_tugboat_get_log($instance_id);
      //watchdog('message', 'regenerating log: ' . count($log));
    }
  }

  return [
    'code' => $status_code,
    'percent' => $percent,
    'message' => $message,
    'log' => $log,
  ];
}

/**
 * Page callback for JSON status message.
 *
 * @param $instance_id
 */
function simplytest_tugboat_status_state_json($instance_id) {
  $state = simplytest_tugboat_status_state($instance_id);
  $state['do_provision'] = $state['code'] === SIMPLYTEST_STATUS_ENQUEUE;
  drupal_json_output($state);
}

/**
 * Get the progress percentage for a running instance.
 *
 * @param $status_code
 *
 * @return string
 */
function simplytest_tugboat_status_running_percentage($status_code) {
  $running_statuses = simplytest_tugboat_status_list('running');
  $total = count($running_statuses);
  $current = 1;
  foreach ($running_statuses as $running_status_code => $status) {
    if ($status_code == $running_status_code) {
      break;
    }
    $current += 1;
  }

  return simplytest_tugboat_percentage($total, $current);
}

/**
 * Formats the percent completion of a progress bar.
 *
 * @param int $total
 *   The total number of operations.
 * @param int $current
 *   The number of the current operation. This may be a floating point number
 *   rather than an integer in the case of a multi-step operation that is not
 *   yet complete; in that case, the fractional part of $current represents the
 *   fraction of the operation that has been completed.
 *
 * @return string
 *   The properly formatted percentage, as a string. We output percentages
 *   using the correct number of decimal places so that we never print "100%"
 *   until we are finished, but we also never print more decimal places than
 *   are meaningful.
 *
 * @see _batch_api_percentage()
 */
function simplytest_tugboat_percentage($total, $current) {
  if (!$total || $total == $current) {
    // If $total doesn't evaluate as true or is equal to the current set, then
    // we're finished, and we can return "100".
    $percentage = "100";
  }
  else {
    // We add a new digit at 200, 2000, etc. (since, for example, 199/200
    // would round up to 100% if we didn't).
    $decimal_places = max(0, floor(log10($total / 2.0)) - 1);
    do {
      // Calculate the percentage to the specified number of decimal places.
      $percentage = sprintf('%01.' . $decimal_places . 'f', round($current / $total * 100, $decimal_places));
      // When $current is an integer, the above calculation will always be
      // correct. However, if $current is a floating point number (in the case
      // of a multi-step batch operation that is not yet complete), $percentage
      // may be erroneously rounded up to 100%. To prevent that, we add one
      // more decimal place and try again.
      $decimal_places++;
    } while ($percentage == '100');
  }
  return $percentage;
}

/**
 * Redirection page callback, to the final sandbox location.
 *
 * @param string $id
 *  The submission id.
 */
function simplytest_tugboat_status_goto($id) {
  // Check for an established URL, forward if exists.
  $url = simplytest_tugboat_url_load($id);
  if (!empty($url) and $url !== '') {
    drupal_goto($url, array('external' => TRUE));
  }

  // Check state.
  $state = simplytest_tugboat_status_state($id);
  if ($state['code'] === FALSE) {
    drupal_set_message(t('The requested submission is not available.'), 'error');
    drupal_goto();
  }
  switch ($state['code']) {
    case SIMPLYTEST_STATUS_ENQUEUE:
    case SIMPLYTEST_STATUS_SPAWNED:
    case SIMPLYTEST_STATUS_PREPARE:
    case SIMPLYTEST_STATUS_DOWNLOAD:
    case SIMPLYTEST_STATUS_PATCHING:
    case SIMPLYTEST_STATUS_INSTALLING:
    case SIMPLYTEST_STATUS_FINALIZE:
      drupal_goto('tugboat/progress/' . $id);
      break;
    case SIMPLYTEST_STATUS_FINISHED:
      $tugboat_url = simplytest_tugboat_url_load($id);
      drupal_goto($tugboat_url, array('external' => TRUE));
      break;
    default:
      drupal_set_message($state['message'], 'error');
      drupal_goto();
  }
}
